
; -------------------------------------------------------
;             PureUnit -- Example code
; -------------------------------------------------------
;
; This is part of the testing code for the PB XML library.
; Not all tests were implemented here, but it shows the 
; idea.

; Test in normal + unicode mode only.
; Thread tests are not needed as the ObjectManager library takes care of all
; thread issues and there are no further global datastructures in this library.
;
PureUnitOptions()

; -------------------------------------------------------
;- Setup
; -------------------------------------------------------

CompilerIf #PB_Compiler_Unicode
  #DefaultEncoding = #PB_Unicode
CompilerElse
  #DefaultEncoding = #PB_Ascii
CompilerEndIf

Enumeration 10 ; leave 0, 1 etc for quick and dirty tests
  #XML_Error
  #XML_Ok
  #XML_NS
EndEnumeration

; re-create some standard XML objects for testing.
; We do it before each test to ensure the same state of the object for each test
; (so we can also do manipulations in the test without problem)
ProcedureUnitBefore before()

  ; Note: be careful when modifying this, as most of the tests rely on
  ;   the structure of this doc
  ;
  Xml$ + "<!DOCTYPE main SYSTEM "+Chr(34)+"uri"+Chr(34)+">" + Chr(10)
  Xml$ + Chr(10)  
  Xml$ + "<main>" + Chr(10)
  Xml$ + "  <sub attribute='abcd'>" + Chr(10)
  Xml$ + "    <empty id='1' />" + Chr(10)
  Xml$ + "    <empty id='2' />" + Chr(10)
  Xml$ + "    <empty id='3' second='' third='bla' />" + Chr(10)
  Xml$ + "  </sub>" + Chr(10)
  Xml$ + "  <!-- comment -->" + Chr(10)
  Xml$ + "  <![CDATA[ cdata content ]]>" + Chr(10)
  Xml$ + "  <?php processing instr ?>" + Chr(10)
  Xml$ + "</main>"
  Assert(CatchXML(#XML_Ok, @Xml$, Len(Xml$)*SizeOf(Character), 0, #DefaultEncoding))  

  ErrorXml$ = "<main>" + Chr(10)
  ErrorXml$ + Chr(10)
  ErrorXml$ + "    <wrong attribute='not complete>" + Chr(10)
  ErrorXml$ + Chr(10)
  ErrprXml$ + "</main>"  
  Assert(CatchXML(#XML_Error, @ErrorXML$, Len(ErrorXML$)*SizeOf(Character), 0, #DefaultEncoding))
  
  XmlNS$ = "<main xmlns:test='http://test.com' >" + Chr(10)
  XmlNS$ + "  <empty xmlns='http://me.com' />" + Chr(10)
  XmlNS$ + "  <test:empty test:A='a' A='b' />" + Chr(10)
  XmlNS$ + "</main>" + Chr(10)  
  Assert(CatchXML(#XML_NS, @XmlNS$, Len(XmlNS$)*SizeOf(Character), 0, #DefaultEncoding))
  
EndProcedureUnit

ProcedureUnitAfter after()
  If IsXML(#XML_Ok) ; so we can call Free() also inside tests
    FreeXML(#XML_Ok)
  EndIf
  
  If IsXML(#XML_Error)
    FreeXML(#XML_Error) 
  EndIf
  
  If IsXML(#XML_NS)
    FreeXML(#XML_NS)
  EndIf
EndProcedureUnit

; -------------------------------------------------------
;- Helper functions
; -------------------------------------------------------

;
; tests if the two nodes are exactly equal
;
Procedure CheckEqualNodes(*Node1, *Node2)

  Assert(XMLNodeType(*Node1)      = XMLNodeType(*Node2))  
  Assert(GetXMLNodeOffset(*Node1) = GetXMLNodeOffset(*Node2))
  Assert(XMLChildCount(*Node1)    = XMLChildCount(*Node2))  
  AssertString(GetXMLNodeText(*Node1), GetXMLNodeText(*Node2))
  
  Type = XMLNodeType(*Node1)
  
  ; Nothing more to check for:
  ;  #PB_XML_Comment
  ;  #PB_XML_CData
  ;  #PB_XML_DTD  
  
  ; Check node name
  ;
  If Type = #PB_XML_Normal Or Type = #PB_XML_Instruction
    AssertString(GetXMLNodeName(*Node1), GetXMLNodeName(*Node2))
  EndIf
  
  ; Check node attributes
  ;
  If Type = #PB_XML_Normal
    Ex1 = ExamineXMLAttributes(*Node1)
    Ex2 = ExamineXMLAttributes(*Node2)
    
    Assert((Ex1 And Ex2) Or (Ex1 = 0 And Ex2 = 0))
    
    If Ex1
      Next1 = NextXMLAttribute(*Node1)
      Next2 = NextXMLAttribute(*Node2)
      
      While Next1        
        Assert(Next2) ; if zero, *Node1 has more attribs than *Node2
        AssertString(XMLAttributeName(*Node1),  XMLAttributeName(*Node2))
        AssertString(XMLAttributeValue(*Node1), XMLAttributeValue(*Node2))        
        Next1 = NextXMLAttribute(*Node1)
        Next2 = NextXMLAttribute(*Node2)
      Wend
      
      Assert(Next2 = 0) ; if nonzero, *Node2 has more attribs than *Node1
    EndIf
  EndIf
  
  ; Check node children
  ;
  If Type = #PB_XML_Root Or Type = #PB_XML_Normal
    *Child1 = ChildXMLNode(*Node1)
    *Child2 = ChildXMLNode(*Node2)
    
    While *Child1
      Assert(*Child2)
      CheckEqualNodes(*Child1, *Child2)
      *Child1 = NextXMLNode(*Child1)
      *Child2 = NextXMLNode(*Child2)
    Wend
    
    Assert(*Child2 = 0)
  EndIf
  
EndProcedure

;
; tests if the two trees are exactly equal
;
Procedure CheckEqualTree(Xml1, Xml2)

  ; check the basic tree stuff
  ;
  Assert(XMLStatus(Xml1)        = XMLStatus(Xml2))
  Assert(GetXMLEncoding(Xml1)   = GetXMLEncoding(Xml2))
  Assert(GetXMLStandalone(Xml1) = GetXMLStandalone(Xml2))

  ; Do a recursive tree compare
  ;
  CheckEqualNodes(RootXMLNode(Xml1), RootXMLNode(Xml2))

  ; Lets export the stuff and compare the markup as well to be sure
  ;
  *Buffer1 = AllocateMemory(ExportXMLSize(Xml1))
  *Buffer2 = AllocateMemory(ExportXMLSize(Xml2))
  Assert(*Buffer1)
  Assert(*Buffer2)    
  ExportXML(Xml1, *Buffer1, ExportXMLSize(Xml1))
  ExportXML(Xml2, *Buffer2, ExportXMLSize(Xml2))  
  
  ;UnitDebug(PeekS(*Buffer1, ExportXMLSize(Xml1)))
  ;UnitDebug(PeekS(*Buffer2, ExportXMLSize(Xml2)))
  
  Assert(ExportXMLSize(Xml1) = ExportXMLSize(Xml2))  
  Assert(CompareMemory(*Buffer1, *Buffer2, ExportXMLSize(Xml1)))  
  
  FreeMemory(*Buffer1)
  FreeMemory(*Buffer2)
EndProcedure

; -------------------------------------------------------
;- Regular command tests
; -------------------------------------------------------

ProcedureUnit Test_CatchXML()
EndProcedureUnit

ProcedureUnit Test_ChildXMLNode()
EndProcedureUnit

ProcedureUnit Test_CopyXMLNode()
  ; copy inside a tree
  ;
  *Old = XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub")
  *New = CopyXMLNode(*Old, MainXMLNode(#XML_Ok), *Old)
  
  Assert(*New = XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub[2]"))
  CheckEqualNodes(*Old, *New)
  
  ; copy to a different tree
  ;
  Assert(CreateXML(0, #PB_UTF8)) ; different encoding should not matter here
  
  *Old = MainXMLNode(#XML_Ok)
  *New = CopyXMLNode(*Old, RootXMLNode(0))
  
  Assert(*New)
  CheckEqualNodes(*Old, *New)
  
  FreeXML(0)  
EndProcedureUnit

ProcedureUnit Test_CreateXML()
EndProcedureUnit

ProcedureUnit Test_CreateXMLNode()
EndProcedureUnit

ProcedureUnit Test_DeleteXMLNode()
EndProcedureUnit

ProcedureUnit Test_ExamineXMLAttributes()
EndProcedureUnit

ProcedureUnit Test_ExportXML()
EndProcedureUnit

ProcedureUnit Test_ExportXMLSize()
EndProcedureUnit

ProcedureUnit Test_FormatXML()
EndProcedureUnit

ProcedureUnit Test_FreeXML()
  FreeXML(#XML_OK)
  Assert(IsXML(#XML_Ok) = 0) 
EndProcedureUnit

ProcedureUnit Test_GetXMLAttribute()
  *Sub = XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub")
  AssertString(GetXMLAttribute(*Sub, "attribute"), "abcd")  
  AssertString(GetXMLAttribute(*Sub, "ATTRIBUTE"), "") ; case sensitive, so should return nothing
  AssertString(GetXMLAttribute(*Sub, "not_existing"), "")  
  AssertString(GetXMLAttribute(ChildXMLNode(*Sub, 1), "id"), "1")
  AssertString(GetXMLAttribute(ChildXMLNode(*Sub, 2), "id"), "2")
  AssertString(GetXMLAttribute(ChildXMLNode(*Sub, 3), "id"), "3")
EndProcedureUnit

ProcedureUnit Test_GetXMLEncoding()
  Assert(GetXMLEncoding(#XML_Ok)    = #DefaultEncoding)
  Assert(GetXMLEncoding(#XML_Error) = #DefaultEncoding)
EndProcedureUnit

ProcedureUnit Test_GetXMLNodeName()
  AssertString(GetXMLNodeName(MainXMLNode(#XML_Ok)), "main")
  AssertString(GetXMLNodeName(ChildXMLNode(MainXMLNode(#XML_Ok), 1)), "sub")
  AssertString(GetXMLNodeName(ChildXMLNode(MainXMLNode(#XML_Ok), 2)), "") ; comments have no name
  AssertString(GetXMLNodeName(ChildXMLNode(MainXMLNode(#XML_Ok), 3)), "") ; cdata have no name
  AssertString(GetXMLNodeName(ChildXMLNode(MainXMLNode(#XML_Ok), 4)), "php") ; processing instructions have one
EndProcedureUnit

ProcedureUnit Test_GetXMLNodeOffset()
  *Sub = XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub")
  
  Assert(GetXMLNodeOffset(ChildXMLNode(*Sub, 1)) = 5) ; newline + 4 spaces
  Assert(GetXMLNodeOffset(ChildXMLNode(*Sub, 2)) = 5)
  Assert(GetXMLNodeOffset(ChildXMLNode(*Sub, 3)) = 5)
  
  Assert(GetXMLNodeOffset(*Sub) = 3) ; newline + 2 spaces  
EndProcedureUnit

ProcedureUnit Test_GetXMLNodeText()
  AssertString(GetXMLNodeText(ChildXMLNode(RootXMLNode(#XML_Ok), 1)), "<!DOCTYPE main SYSTEM "+Chr(34)+"uri"+Chr(34)+">")
  AssertString(GetXMLNodeText(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub/empty")), "")
  AssertString(GetXMLNodeText(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#comment")), " comment ")
  AssertString(GetXMLNodeText(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#cdata")), " cdata content ")
  AssertString(GetXMLNodeText(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#instruction")), "processing instr ") ; leading whitespace is cut here!
  AssertString(GetXMLNodeText(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub")), Chr(10)+"    "+Chr(10)+"    "+Chr(10)+"    "+Chr(10)+"  ")
EndProcedureUnit

ProcedureUnit Test_GetXMLStandalone()
  CompilerIf #PB_Compiler_Unicode
    Xml1$ = "<?xml version='1.0' encoding='UTF-16' standalone='yes' ?><main/>"
    Xml2$ = "<?xml version='1.0' encoding='UTF-16' standalone='no' ?><main/>"  
    Xml3$ = "<?xml version='1.0' encoding='UTF-16' ?><main/>"  
  CompilerElse
    Xml1$ = "<?xml version='1.0' encoding='ISO-8859-1' standalone='yes' ?><main/>"
    Xml2$ = "<?xml version='1.0' encoding='ISO-8859-1' standalone='no' ?><main/>"
    Xml3$ = "<?xml version='1.0' encoding='ISO-8859-1' ?><main/>"
  CompilerEndIf  
  
  Assert(CatchXML(0, @Xml1$, Len(Xml1$)*SizeOf(Character)))
  Assert(CatchXML(1, @Xml2$, Len(Xml2$)*SizeOf(Character)))
  Assert(CatchXML(2, @Xml3$, Len(Xml3$)*SizeOf(Character)))
  Assert(CreateXML(3))

  Assert(GetXMLStandalone(0) = #PB_XML_StandaloneYes)
  Assert(GetXMLStandalone(1) = #PB_XML_StandaloneNo)
  Assert(GetXMLStandalone(2) = #PB_XML_StandaloneUnset)
  Assert(GetXMLStandalone(3) = #PB_XML_StandaloneUnset)
  
  Assert(GetXMLStandalone(#XML_OK) = #PB_XML_StandaloneUnset)
  
  FreeXML(0)
  FreeXML(1)   
  FreeXML(2)   
  FreeXML(3)   
EndProcedureUnit

ProcedureUnit Test_IsXML()
  Assert(IsXML(#XML_Error))
  Assert(IsXML(#XML_OK))
  Assert(IsXML(12345) = 0)
  Assert(IsXML(-1) = 0)
EndProcedureUnit

ProcedureUnit Test_LoadXML()
EndProcedureUnit

ProcedureUnit Test_MainXMLNode()
  Assert(MainXMLNode(#XML_Ok))
  
  Assert(CreateXML(0))
  Assert(MainXMLNode(0) = 0)
  FreeXML(0)
EndProcedureUnit


ProcedureUnit Test_NextXMLAttribute()
  ; tested in Test_ExamineXMLAttributes()
EndProcedureUnit

ProcedureUnit Test_NextXMLNode()
EndProcedureUnit

ProcedureUnit Test_ParentXMLNode()
EndProcedureUnit

ProcedureUnit Test_PreviousXMLNode()
EndProcedureUnit

ProcedureUnit Test_RemoveXMLAttribute()
EndProcedureUnit

ProcedureUnit Test_ResolveXMLAttributeName()
  *Node = XMLNodeFromPath(RootXMLNode(#XML_NS), "/main/test:empty")
  AssertString(ResolveXMLAttributeName(*Node, "test:A"), "http://test.com/A")
  AssertString(ResolveXMLAttributeName(*Node, "test:A", "#"), "http://test.com#A")
  AssertString(ResolveXMLAttributeName(*Node, "A"), "A") ; no default namespacing here!! (XML spec)
EndProcedureUnit

ProcedureUnit Test_ResolveXMLNodeName()
  AssertString(ResolveXMLNodeName(MainXMLNode(#XML_NS)), "main")
  AssertString(ResolveXMLNodeName(XMLNodeFromPath(RootXMLNode(#XML_NS), "/main/empty")), "http://me.com/empty")
  AssertString(ResolveXMLNodeName(XMLNodeFromPath(RootXMLNode(#XML_NS), "/main/empty"), "#"), "http://me.com#empty")
  AssertString(ResolveXMLNodeName(XMLNodeFromPath(RootXMLNode(#XML_NS), "/main/test:empty")), "http://test.com/empty")
EndProcedureUnit

ProcedureUnit Test_RootXMLNode()
  Assert(RootXMLNode(#XML_Ok))    ; may never be 0
  Assert(RootXMLNode(#XML_Error)) ; may never be 0
EndProcedureUnit

ProcedureUnit Test_SaveXML()
EndProcedureUnit

ProcedureUnit Test_SetXMLAttribute()
EndProcedureUnit

ProcedureUnit Test_SetXMLEncoding()
EndProcedureUnit

ProcedureUnit Test_SetXMLNodeName()
EndProcedureUnit

ProcedureUnit Test_SetXMLNodeOffset()
EndProcedureUnit

ProcedureUnit Test_SetXMLNodeText()
EndProcedureUnit

ProcedureUnit Test_SetXMLStandalone()
EndProcedureUnit

ProcedureUnit Test_XMLAttributeName()
  ; tested in Test_ExamineXMLAttributes()
EndProcedureUnit

ProcedureUnit Test_XMLAttributeValue()
  ; tested in Test_ExamineXMLAttributes()
EndProcedureUnit

ProcedureUnit Test_XMLChildCount()
  Assert(XMLChildCount(RootXMLNode(#XML_Ok)) = 2)
  Assert(XMLChildCount(MainXMLNode(#XML_Ok)) = 4)
  Assert(XMLChildCount(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/#dtd")) = 0)
  Assert(XMLChildCount(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub")) = 3)
  Assert(XMLChildCount(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#comment")) = 0)
  Assert(XMLChildCount(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#cdata")) = 0)
  Assert(XMLChildCount(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#instruction")) = 0)
EndProcedureUnit

ProcedureUnit Test_XMLError()
  AssertString(XMLError(#XML_Ok), "")
  AssertString(XMLError(#XML_Error), "unclosed token")
EndProcedureUnit

ProcedureUnit Test_XMLErrorLine()
  Assert(XMLErrorLine(#XML_Error) = 3)
EndProcedureUnit

ProcedureUnit Test_XMLErrorPosition()
  Assert(XMLErrorPosition(#XML_Error) = 5)
EndProcedureUnit


ProcedureUnit Test_XMLNodeFromPath()
  *Node = XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/sub/empty[2]")
  Assert(*Node)
  Assert(*Node = XMLNodeFromPath(*Node, "/main/*/*[2]"))
  Assert(*Node = NextXMLNode(XMLNodeFromPath(MainXMLNode(#XML_Ok), "sub/empty")))
EndProcedureUnit


ProcedureUnit Test_XMLNodeType()
  Assert(XMLNodeType(RootXMLNode(#XML_Ok)) = #PB_XML_Root)
  Assert(XMLNodeType(MainXMLNode(#XML_Ok)) = #PB_XML_Normal)
  Assert(XMLNodeType(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/#dtd"))              = #PB_XML_DTD)
  Assert(XMLNodeType(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#comment"))     = #PB_XML_Comment)
  Assert(XMLNodeType(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#cdata"))       = #PB_XML_CData)
  Assert(XMLNodeType(XMLNodeFromPath(RootXMLNode(#XML_Ok), "/main/#instruction")) = #PB_XML_Instruction)
EndProcedureUnit

ProcedureUnit Test_XMLStatus()
  Assert(XMLStatus(#XML_OK) = #PB_XML_Success)
  Assert(XMLStatus(#XML_Error) = #PB_XML_UnclosedToken)
EndProcedureUnit

